// SPDX-FileCopyrightText: tuberry
// SPDX-License-Identifier: GPL-3.0-or-later

import St from 'gi://St';
import Gio from 'gi://Gio';
import GObject from 'gi://GObject';

import {EventEmitter} from 'resource:///org/gnome/shell/misc/signals.js';
import {loadInterfaceXML} from 'resource:///org/gnome/shell/misc/fileUtils.js';
import {TransientSignalHolder} from 'resource:///org/gnome/shell/misc/signalTracker.js';
import {Extension as ExtensionBase, gettext as _} from 'resource:///org/gnome/shell/extensions/extension.js';

import {hook, xnor} from './util.js';

export {_};
export const myself = () => Extension.lookupByURL(import.meta.url); // NOTE: https://github.com/tc39/proposal-json-modules
export const debug = (...xs) => console.debug(`[${myself().uuid}]`, ...xs);
export const ruin = o => o && (o.destroy ?? o.run_dispose)?.bind(o)();
export const omit = (o, ...ks) => ks.forEach(k => { ruin(o[k]); o[k] = null; });
// TODO: need https://gitlab.gnome.org/GNOME/gnome-shell/-/merge_requests/2542
const onus = o => [o, o.$hub].find(x => GObject.type_is_a(x, GObject.Object) && GObject.signal_lookup('destroy', x)) ??
    (() => { throw Error('undestroyable'); })(); // NOTE: https://github.com/tc39/proposal-throw-expressions#todo
export const connect = (tracker, ...args) => (x => args.forEach(([emitter, ...argv]) => emitter.connectObject(...argv, x)))(onus(tracker));
export const disconnect = (tracker, ...args) => (x => args.forEach(emitter => emitter?.disconnectObject(x)))(onus(tracker));
export const open = uri => Gio.AppInfo.launch_default_for_uri(uri, global.create_app_launch_context(0, -1));
export const copy = (text, primary) => St.Clipboard.get_default().set_text(primary ? St.ClipboardType.PRIMARY : St.ClipboardType.CLIPBOARD, text);
export const paste = primary => new Promise(resolve => St.Clipboard.get_default().get_text(
    primary ? St.ClipboardType.PRIMARY : St.ClipboardType.CLIPBOARD, (_c, x) => x && resolve(x)));

export class Mortal extends EventEmitter {
    $hub = new TransientSignalHolder(this);

    destroy() {
        this.emit('destroy');
        omit(this, '$hub');
    }
}

/**
 * @template T
 * @param {T} doom
 * @return {T}
 */
export function degrade(doom, host) {
    host.connect('destroy', () => omit(doom, ...Object.keys(doom)));
    return doom;
}

export function light(callback) {
    let ColorProxy = Gio.DBusProxy.makeProxyWrapper(loadInterfaceXML('org.gnome.SettingsDaemon.Color'));
    return hook({'g-properties-changed': (a, p) => { if(p.lookup_value('NightLightActive', null)) callback(a.NightLightActive); }},
        new ColorProxy(Gio.DBus.session, 'org.gnome.SettingsDaemon.Color', '/org/gnome/SettingsDaemon/Color', a => callback(a.NightLightActive)));
}

export class Extension extends ExtensionBase {
    enable() {
        this.$hub = new this.$klass(this.getSettings());
    }

    disable() {
        omit(this, '$hub');
    }
}

export class Source {
    constructor(summon, dispel = ruin, enable, ...args) {
        this.summon = (...xs) => (this.hub = summon(...xs));
        this.dispel = () => { dispel(this.hub); this.hub = null; };
        this.revive = (...xs) => { this.dispel(); return this.summon(...xs); };
        this.toggle = (a, ...xs) => xnor(a, this.hub) || (a ? this.summon(...xs) : this.dispel());
        if(enable) this.summon(...args);
        this.destroy = this.dispel;
    }
}

export class Setting {
    #map = new WeakMap();
    constructor(prop, gset, obj, cluster) {
        this.gset = typeof gset === 'string' ? new Gio.Settings({schema: gset}) : gset;
        this.attach(prop, obj, cluster);
    }

    get(prop, obj) {
        return (([key, type]) => this.gset[`get_${type}`](key))(this.#map.get(obj)[prop]);
    }

    set(prop, value, obj) {
        (([key, type]) => this.gset[`set_${type}`](key, value))(this.#map.get(obj)[prop]);
    }

    attach(props, obj, cluster) { // cluster && props <- { prop: [key, type, output] }
        this.#map.has(obj) ? Object.assign(this.#map.get(obj), props) : this.#map.set(obj, props);
        let callback = cluster ? x => { obj[cluster] = [x, this.get(x, obj), this.#map.get(obj)[x][2]]; } : x => { obj[x] = this.get(x, obj); };
        Object.entries(props).forEach(([k, [x]]) => { callback(k); connect(obj, [this.gset, `changed::${x}`, () => callback(k)]); });
        return this;
    }

    detach(obj) {
        if(this.#map.has(obj)) disconnect(obj, this.gset);
    }
}
